import logging

import git

from .. import line_range

REMOTE = "origin"
URL_REPLACEMENTS = {"git@github.com:": "https://github.com/"}


class RepoApi(object):
    """wrapper class for git.Repo"""

    def __init__(self, repo: git.Repo) -> None:
        self._repo = repo

    @property
    def current_branch(self) -> str:
        """Return the current local git branch"""
        return str(self._repo.active_branch)

    @property
    def github_url(self) -> str:
        """Convert an https or ssh remote into a github url"""
        remote_url = self._repo.remotes[REMOTE].url
        # if we have an ssh remote, convert it to a URL
        for ssh_login in URL_REPLACEMENTS:
            if remote_url.startswith(ssh_login):
                logging.debug(f"Remote URL starts with ssh_login: {remote_url}")
                remote_url = remote_url.lstrip(ssh_login)
                remote_url = URL_REPLACEMENTS[ssh_login] + remote_url
        remote_url = remote_url.rstrip(".git")
        return remote_url

    def get_permalink(
        self, path_from_git_root: str, range: line_range.LineRange
    ) -> str:
        """Create a permalink for the editor's selection"""
        return (
            f"{self.github_url}/blob/{self.current_branch}/{path_from_git_root}#{range}"
        )



#! python
import logging
import os
import subprocess
import tempfile
from typing import Optional

import click
import git

from . import os as kak_plugins_os
from .apis import clipboard
from .apis import git as git_api
from .apis import kak

DEFAULT_LOGFILE = "github-permalink.log"

LOG_LEVELS = {
    "info": logging.INFO,
    "debug": logging.DEBUG,
    "critical": logging.CRITICAL,
}


@click.command()
@click.option(
    "-c",
    "--clipboard-command",
    envvar="CLIPBOARD",
    help="Program that is used to write to the system clipboard. "
    + "CLIPBOARD will be read from the environment if this flag is not specified.",
)
@click.option(
    "-l",
    "--log-level",
    default="info",
    help="Level of information to write to the logs.",
    type=click.Choice(LOG_LEVELS),
    show_default=True,
)
@click.option(
    "-w",
    "--write-to-logfile",
    is_flag=True,
    help="If this flag is on, write to a logfile instead.",
)
@click.option("-p", "--log-path", help="Where to write the logfile.")
def main(
    log_level: str,
    write_to_logfile: bool,
    log_path: Optional[str],
    clipboard_command: Optional[str],
) -> None:  # pragma: no cover
    """Call kcr to get editor info, then parse it and write it to the clipboard"""
    # TODO: write click tests. maybe need to use a mock?
    if not clipboard_command:
        raise RuntimeError("CLIPBOARD is not set")
    if log_path:
        # if a user specifies a log path, we can assume they want to log
        write_to_logfile = True
    _setup_logging(log_level, write_to_logfile, log_path)
    get_github_permalink(clipboard_command)
    logging.info("=== script complete ===")


def _setup_logging(
    log_level: str, write_to_logfile: bool, log_path: Optional[str]
) -> None:  # pragma: no cover
    if write_to_logfile:
        if log_path is None:
            temp_dir = tempfile.gettempdir()
            log_path = f"{temp_dir}/{DEFAULT_LOGFILE}"
        logging.basicConfig(
            filename=log_path,
            format="%(levelname)s: %(message)s",
            level=LOG_LEVELS[log_level],
        )
    else:
        # write to stderr
        logging.basicConfig(
            format="%(levelname)s: %(message)s",
            level=LOG_LEVELS[log_level],
        )
    logging.info(f"log level set to {log_level.upper()}")


def get_github_permalink(clipboard_command: str) -> None:  # pragma: no cover
    # TODO: break this file up and stop ignoring coverage
    absolute_path, selection_desc = kak.kcr_get(
        subprocess.run, ["buffile", "selection_desc"]
    )
    selection = kak.SelectionDescription(selection_desc)
    git_root = kak_plugins_os.get_git_root(os.path.exists, absolute_path)
    repo = git_api.RepoApi(git.Repo(git_root))
    relative_path = os.path.relpath(absolute_path, git_root)
    permalink = repo.get_permalink(relative_path, selection.range)
    clipboard_job = clipboard.ClipboardJob(clipboard_command, permalink)
    clipboard.write_to_clipboard(subprocess.run, clipboard_job)



import os
from collections.abc import Callable


def get_git_root(path_exists_function: Callable, current_path: str) -> str:
    """Find the absolute path of the given path's git root"""
    # Traverse up the directory tree until we find a directory
    # that contains a .git directory
    while not path_exists_function(f"{current_path}/.git"):
        if current_path == "/":
            raise RuntimeError(f"{current_path} is not in a git directory")
        current_path = os.path.dirname(current_path)
    return current_path



import collections
import logging
from collections.abc import Callable

# represents a unit of work for writing to a clipboard
ClipboardJob = collections.namedtuple("ClipboardJob", ["clipboard_command", "message"])


def write_to_clipboard(runner: Callable, job: ClipboardJob) -> None:
    """Make a system call to write to a clipboard.

    args:
        - runner: a callable like subprocess.run that will make a system call
        - job: the unit of work to perform
    """
    logging.debug(f"writing '{job.message}' to {job.clipboard_command}")
    runner(
        job.clipboard_command, input=job.message, text=True, check=True
    )  # noqa: S603



import json
import logging
from collections.abc import Callable
from typing import List

from .. import line_range


def kcr_get(runner: Callable, values: List) -> List:
    kcr_command = ["kcr", "get"]
    for value in values:
        kcr_command.append("--value")
        kcr_command.append(value)
    logging.debug(f"running commmand: '{kcr_command}'")
    result = runner(kcr_command, capture_output=True, check=True)
    kcr_output = str(result.stdout, encoding="utf-8").strip()
    logging.debug(f"kcr output: {kcr_output}")
    return json.loads(kcr_output)


class SelectionDescription(object):
    def __init__(self, selection_desc: str) -> None:
        """Parse a selction description from kakoune"""
        anchor_pos, cursor_pos = selection_desc.split(",")
        anchor_line = int(anchor_pos.split(".")[0])
        cursor_line = int(cursor_pos.split(".")[0])
        logging.debug(f"anchor at {anchor_line}; cursor at {cursor_line}")
        self._line_range = line_range.LineRange(
            min(anchor_line, cursor_line), max(anchor_line, cursor_line)
        )

    @property
    def range(self) -> line_range.LineRange:
        return self._line_range



class LineRange(object):
    """Describes a range of lines, both in Kakoune and Github"""

    def __init__(self, start: int, stop: int) -> None:
        if stop < start:
            # LineRanges are dumb objects and should always have start/stop ordered
            raise ValueError(f"Invalid range. {stop} < {start}")
        if start < 1:
            # kakoune and github start their line counts at 1
            raise ValueError(f"Invalid range start. {start} < 1")
        self._start = start
        self._stop = stop

    def __str__(self) -> str:
        if self._start == self._stop:
            return f"L{self._start}"
        else:
            return f"L{self._start}-L{self._stop}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LineRange):
            return (self._start, self._stop) == (other._start, other._stop)
        return False



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kak_plugins', 'kak_plugins.apis']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.14,<4.0.0', 'click>=7.1.2,<8.0.0']

entry_points = \
{'console_scripts': ['github-permalink = kak_plugins.github_permalink:main']}

setup_kwargs = {
    'name': 'kak-plugins',
    'version': '0.4.2',
    'description': '',
    'long_description': '[![Tests](https://github.com/abstractlyZach/kak_plugins/workflows/Tests/badge.svg)](https://github.com/abstractlyZach/kak_plugins/actions?workflow=Tests)\n[![PyPI](https://img.shields.io/pypi/v/kak-plugins.svg)](https://pypi.org/project/kak-plugins/)\n[![Codecov](https://codecov.io/gh/abstractlyZach/kak_plugins/branch/main/graph/badge.svg)](https://codecov.io/gh/abstractlyZach/kak_plugins)\n\n\n# Zach\'s Overengineered Kakoune Plugins\nI loved seeing the plugins that people write for [Kakoune](http://kakoune.org/), so I decided to write some on my own, but with extreme gusto. Sure, you could probaly accomplish the same results with a string of 7 shell commands, but where\'s the fun in that when you could have rigorous testing, fancy documentation, interactive menus and debugging, and powerful continuous integration?!\n\n## Installation\nI recommend using [pipx](https://pipxproject.github.io/pipx/installation/) for installation. It allows you to install python packages on your machine in separate virtual environments without having to manage the virtual environments yourself. `pip` also works if you prefer that.\n```\npipx install kak_plugins\n```\n\n## Dependencies\n* [Kakoune](http://kakoune.org/) of course\n* [kakoune.cr](https://github.com/alexherbo2/kakoune.cr)\n    * enables us to retrieve info from Kakoune\n    * provides an interface to control Kakoune\n* A clipboard command-line utility. I use these:\n    * `pbcopy` for OSX\n    * [xclip](https://github.com/astrand/xclip) for Linux\n    * [wl-clipboard](https://github.com/bugaevc/wl-clipboard) for Wayland (if you don\'t know what this is and you use Linux, you\'ll probably use `xclip`)\n\n\n## Setup\nThere are some environment varibles you will need to define in order to use these plugins. You would probably define these in your `~/.bashrc`, `zshrc`, or `~/.profile`. I define mine [here](https://github.com/abstractlyZach/dotfiles/blob/master/common/.profile]\n```\n# program that reads stdin and writes to your system clipboard\nexport CLIPBOARD="pbcopy"\n```\n\n## Logging\nLogs are a good way of getting an idea of what\'s going on in the code. Logs will be written to stderr by default, but if you turn on logging and don\'t specify a path, then logs will be written in a temporary directory in a file that has the name of the plugin. See here for rules on [where you can find logs on your system](https://docs.python.org/3/library/tempfile.html#tempfile.gettempdir)\n\n# Plugins\n\n## github-permalink\nCreate a permalink to a line or range of lines in a GitHub repo that matches your current selection in Kakoune. Then copy that permalink to your clipboard program.\n\n### in kak\nThis method is great for using in your everyday editing\n\n1. open a file in Kakoune\n1. make a selection\n1. in normal mode, use `:$ github-permalink`\n1. you now have a permalink to your kakoune selection. it should look something like this https://github.com/abstractlyZach/kak_plugins/blob/write-readme/README.md#L40\n\nI like [binding this command](https://github.com/abstractlyZach/dotfiles/blob/master/kak/kakrc#L12) to hotkeys so I can hit 2 buttons and then paste the link into Slack or something.\n\n### in a terminal\nThis method is great for learning, development, and debugging\n\n1. open a file in kakoune\n1. make a selection\n1. open a connected terminal. there are a couple of recommended methods\n    * use `:>` in normal mode\n    * [kcr-fzf-shell](https://github.com/alexherbo2/kakoune.cr/blob/master/share/kcr/commands/fzf/kcr-fzf-shell)\n1. `github-permalink --help`\n',
    'author': 'abstractlyZach',
    'author_email': 'zach3lee@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/abstractlyZach/kak_plugins',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
